#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#
# Copyright (c) 2017 Stephen Bunn (stephen@bunn.io)
# MIT License <https://opensource.org/licenses/MIT>

from .lower import (LowerRuleTest,)
from .upper import (UpperRuleTest,)
from .capitalize import (CapitalizeRuleTest,)
from .title import (TitleRuleTest,)
from .lstrip import (LstripRuleTest,)
from .rstrip import (RstripRuleTest,)
from .strip import (StripRuleTest,)
from .increment import (IncrementRuleTest,)
from .decrement import (DecrementRuleTest,)
from .substitute import (SubstituteRuleTest,)
from .translate_text import (TranslateTextRuleTest,)
from .translate_date import (TranslateDateRuleTest,)

from .add_column import (AddColumnRuleTest,)
from .remove_column import (RemoveColumnRuleTest,)
